/*
 * This file is part of the "OCAPI" project, a C implementation of the
 * OPeNDAP Data Access Protocol.
 *
 * Copyright (c) 2005 OPeNDAP, Inc.
 * Author: John Chamberlain <jchamberlain at opendap.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 */

#include <stdio.h>

/* #define unsigned */

/*! \brief Returned result code.

  All the OCAPI function calls return this enum.
 */
typedef enum { FAILURE, SUCCESS } eRESULT;

/*! May be used to track a connection's status. */
typedef enum {
    CS_Uninitialized,
    CS_Initial,
    CS_Opening,
    CS_Reading,
    CS_Full,
    CS_Complete
} eCONNECTION_STATE;

/*! OPeNDAP clients can make several different kinds of requests to a
  server.  A client can request raw data, a catalog, a DDS, a DAS, and
  so on.  (Consult the OPeNDAP User Guide for more information about
  these.)  This enum is used to specify which kind of request is being
  made in a \c URL structure.
 */
typedef enum {
    RC_Raw,
    RC_Catalog,
    RC_Directory,
    RC_Data,
    RC_DDX,
    RC_DDS,
    RC_DAS,
    RC_Info
} eREQUEST_CLASS;

/*! Specifies how detailed the logging of data connections is to be. */
typedef enum {
    LL_NoChange,
    LL_Brief,
    LL_Normal,
    LL_All
} eLOG_LEVEL;

/*! \brief Specifies a proxy host.

  If your connection to the internet requires a proxy server, you
  can specify it in the \c ConnectionParameters structure, using this
  \c Proxy structure. */
typedef struct {
    char *sHost;
    int iPort;
    char *sUser;
    char *sPassword;
} Proxy;

/*! \brief Contains a parsed URL.

  An OPeNDAP URL consists of several pieces: a server name, a path,
  a query, and so forth.  Rather than store them all in one
  character string, OCAPI provides a data structure for holding all
  the pieces together.
 */
typedef struct {
    /*! The unparsed URL string.  Used for reference only. */
    char *sRawURL;
    /*! The host name from the URL. */
    char *sHost;
    /*! The port to be used.  This defaults to 80. */
    int iPort;
    /*! The pathname of the data file from the OPeNDAP URL. */
    char *sPath;
    /*! If there is a query expression, it's enclosed in this string. */
    char *sQuery;
    /*! The request class.  An OPeNDAP URL can be used to retrieve
       several different kinds of information. This enum specifies which
       one is desired.  The default is \c RC_Info. */
    eREQUEST_CLASS eRequestClass;
    char *sCanonicalURL;	/* the official form of the string used to get the data */
} URL;

typedef struct {
/*! \brief Describes a connection to be made to an OPeNDAP data server.

  The connection between a client and an OPeNDAP data server can be
  characterized by several parameters besides the URL.  For
  convenience, they are collected in this structure.

  There are default values for most of these parameters, documented
  at the \c configure function.
*/
    /*! If your connection requires a proxy, it can be specified
       here. */
    Proxy *pstructProxy;
    /*! The \c URL structure containing the address (and any query
       expression) for the OPeNDAP data desired. */
    URL *pstructURL;
    /*! The number of seconds to wait before timing out on a connect.
       -1 = no change, 0 = infinite. */
    int iConnectTimeout_seconds;
    /* The user name for HTTP requests*/
    char *sUserC;
    /*The password for HTTP requests*/
     char *sPasswordC;
    /*! The number of seconds to wait before timing out on a data read.
       -1 = no change, 0 = infinite. */
    int iReadTimeout_seconds;
    /*! The connection protocol to use. */
    char *sProtocol;
    /*! The name of the UserAgent, supplied by the OCAPI client to the
       data server. */
    char *sUserAgent;
    /*! Data that belongs in the client header supplied to the server
       should be put here. There are no default values supplied for this
       parameter. */
    char *sClientHeader;
    /*! If logging messages should be directed to a particular stream
       supply it here, otherwise there will be no logging. */
    FILE *pLoggingStream;
    /*If a file name is provided for cookies to be used*/
    char *cook;
    /*If want to shut off ssl verification*/
    int sVerify;
    /*used to allow for compression*/
    char *sCompress;

} ConnectionParameters;

typedef struct {
/*! \brief Stores any values that persist across calls within the same thread.

  To avoid the use of internal globals and to allow multiple threads of a client
  program to simultaneously use the OCAPI this structure stores values that will
  persist across multiple requests made by the same thread. Each thread that uses
  the OCAPI must store a pointer to this structure and supply when necessary.

  The memory usage by members of the structure is managed within the OCAPI and
  the client does not need to allocate memory for this structure or free it.
*/
    /*! Currently this is the only stored member. */
    ConnectionParameters *pConnectionParameters;
} ThreadSpecificStorage;

/*! \brief Holds conversion formats.

  Use this structure to hold a selection of data conversion
  formats. */
typedef struct {
    int ctFormats;
    char **asTitles;		/* dynamic array of titles and */
    char **asDescriptions;	/*    descriptions */
} ConversionFormats;

#define DBOOLEAN int
#define DBYTE unsigned char
#define DINT16 short
#define DUINT16 unsigned short
#define DINT32 int
#define DUINT32 unsigned int
#define DFLOAT32 float
#define DFLOAT64 double

#define MAX_Dimension_Length 2147483647	/* 2^31 - 1, the size of a signed int minus one for one-based arrays */

/*! \brief Defines a format for storing structures

  Defines a format for storing structures
*/
typedef enum {
    FORMAT_OPeNDAPBinary,
    FORMAT_HDF,
    FORMAT_NetCDF
} eFORMATS;

/*! \brief Specifies the kind of data in an OPeNDAP node tree.

  An OPeNDAP node tree can contain any one of four different
  structures:

  - The DDS holds only structural metadata (array dimensions,
    structure members, and so on).
  - The DataDDS holds data, as well as structural metadata.
  - The DAS holds only attribute information.
  - The DDX, used by some OPeNDAP servers, holds data,
    structural metadata and attributes.

  See the OPeNDAP user manual for a detailed look at the different
  OPeNDAP data structures.
*/
typedef enum {
    STRUCTURE_undefined,
    STRUCTURE_DataDDS,
    STRUCTURE_DDX,
    STRUCTURE_DDS,
    STRUCTURE_DAS
} eSTRUCTURE_TYPE;

/*! \brief Defines the type of an array.

  This enum is used with \c NodeContent_Array to specify the kind of
  data held by that array.
*/
typedef enum {
    ARRAY_Byte,
    ARRAY_Int16,
    ARRAY_UInt16,
    ARRAY_Int32,
    ARRAY_UInt32,
    ARRAY_Float32,
    ARRAY_Float64,
    ARRAY_String,
    ARRAY_URL,
    ARRAY_Sequence,
    ARRAY_Array,
    ARRAY_Grid,
    ARRAY_Structure
} eARRAY_TYPE;

/*! \brief Identifies the node data type.

  This flag indicates what kind of data is attached to a node in the
  OPeNDAP tree structure. */
typedef enum {
    NODE_root = 0x20DE0001,
    NODE_Byte = 0x20DE0002,
    NODE_Int16 = 0x20DE0003,
    NODE_UInt16 = 0x20DE0004,
    NODE_Int32 = 0x20DE0005,
    NODE_UInt32 = 0x20DE0006,
    NODE_Float32 = 0x20DE0007,
    NODE_Float64 = 0x20DE0008,
    NODE_String = 0x20DE0009,
    NODE_URL = 0x20DE000A,
    /* get rid of list*/
    NODE_List = 0x20DE000B,
    NODE_Sequence = 0x20DE000C,
    NODE_Array = 0x20DE000D,
    NODE_Grid = 0x20DE000E,
    NODE_Structure = 0x20DE000F
} eNODE_TYPE;

/*! \brief Identifies the data type of an attribute value. */
typedef enum {
    ATTRIBUTE_unknown = 0x0A770001,
    ATTRIBUTE_alias = 0x0A770002,
    ATTRIBUTE_Byte = 0x0A770003,
    ATTRIBUTE_Int16 = 0x0A770004,
    ATTRIBUTE_UInt16 = 0x0A770005,
    ATTRIBUTE_Int32 = 0x0A770006,
    ATTRIBUTE_UInt32 = 0x0A770007,
    ATTRIBUTE_Float32 = 0x0A770008,
    ATTRIBUTE_Float64 = 0x0A770009,
    ATTRIBUTE_String = 0x0A77000A,
    ATTRIBUTE_URL = 0x0A77000B
} eATTRIBUTE_TYPE;

/*! \brief Contains a string.
  This data structure actually contains the data for an OPeNDAP
  String.
 */
typedef struct {
    unsigned int iLength;
    char *sContent;		/* will be null terminated */
} NodeContent_String;

/*! \brief Contains an Array

  This data structure contains an OPeNDAP Array.  An Array consists of
  the following pieces:

  - A name for the Array
  - A list of dimension sizes
  - A list of dimension names
  - The Array data
 */
typedef struct {
    eARRAY_TYPE eType;
    /*! The number of array dimensions. */
    unsigned int ctDimensions;
    /*! The length of each dimension of the array. The vector is one-based. */
    unsigned int *iDimensionLengthArray1;
    /*! The names of each dimension of the array. The vector is one-based. */
    char **sDimensionName1;
    /*! The name of the Array. */
    char *sArrayName;
    /*! The Array data.  Stored in row-major order. */
    char *array;
} NodeContent_Array;

/*! \brief Contains an OPeNDAP Grid.

  A Grid contains an Array, and a number of "map vectors" used to
  define the use of each dimension of the main array.  For example, a
  Grid might contain a data array that has named dimensions \c lat and
  \c lon, along with map vectors \c lat and \c lon.  The values in the
  map vectors defines where each row and column of the main array is
  located.
 */
typedef struct {
    /*! The number of map vectors. */
    unsigned int ctMappings;
    /*! The Grid data array. */
    NodeContent_Array *values;
    /*! The Grid map vectors. */
    NodeContent_Array **maps1;	/* one-based */
} NodeContent_Grid;

/*! \brief Contains an OPeNDAP Sequence.

  An OPeNDAP Sequence is essentially the same as a Structure repeated
  one or more times.  You can think of it as a relational database
  table, and this is the OPeNDAP data type used to serve data from
  database tables.
 */

typedef struct {
    /*! The number of rows in the Sequence. */
    unsigned int ctRows;
    /*! The number of different fields in each row of the Sequence. */
    unsigned int ctFields;
    /*! An array of field types.
       This is a pointer to a zero-based array containing the field
       types, and having \c ctFields entries. */
    eNODE_TYPE *aeFieldType;
    /*! The data in the Sequence.
       This is a zero based array of vectors (one-dimensional arrays).
       Each vector contains all the rows of for each field allocated as a
       contiguous block of memory.  Each entry in the field vectors is
       either the value itself if it is a simple numerical type, or a
       a pointer to the node content type if it is a String, Grid, Array,
       Sequence or Structure */
    void **aFieldValue;
} NodeContent_Sequence;

/*! \brief Contains a Structure contained in a Sequence.

  This data type is used when a Structure appears as a field in a
  Sequence.  Structures that are not contained in a Sequence are
  nodes of the DAPObject tree and their content nodes are NULL.
 */
typedef struct {
    /*! The number of members of the Structure. */
    unsigned int ctMembers;
    /*! An array of node types, one for each Structure member.
       These may be any \c eNODE_TYPE except \c root.
     */
    eNODE_TYPE *array_eNodeType;
    /*! An array of Structure members.
       This array contains a pointer to an \c array_NodeContent structure
       for each Structure member.
     */
    void **array_NodeContent;
} NodeContent_Structure;

/*! \brief Holds an attribute for some data value.

  This structure is used to hold an attribute for some data value.  An
  attribute may be a simple scalar value or may be a vector of
  values.  But a single attribute has a single type, even if it is a
  vector.
 */
typedef struct {
    /*! The type of the attribute. */
    eATTRIBUTE_TYPE eType;
    /*! An indeterminate length string containing the name of the attribute. */
    char *sName;
    /*! The number of values in this attribute. */
    unsigned int ctValues;
    /*! A pointer to a list of pointers to each value of the attribute.
       Note that a single valued attribute is stored in an array of length one.
     */
    void **pValueList;
} DAPAttribute;

/*! \brief A list of attributes
  This is used as a container for a collection of attributes.  A
  data value or set of values can be said to have several attributes.
  They are all collected in one DAPAttributeTable, which is attached to
  the tree node for that data value.
*/
typedef struct {
    /*! The number of attributes in this list. */
    unsigned int ctAttributes;
    /*! A pointer to a list of pointers to each attribute name. */
    char **pNameList;
    /*! A pointer to a list of pointers to each attribute. */
    DAPAttribute **pAttributeList;
} DAPAttributeTable;

/*! \brief The basic building block of the OPeNDAP tree structure.

  The tree structure that holds an OPeNDAP object is composed of
  nodes using this data structure.  (See \ref sec-data.)  The
  data in the structure consists of:

  - A data type;
  - A name;
  - An estimated size;
  - A set of attributes; and
  - A pointer to its own value (or values).

  In addition, the structure contains a pointer to its parent node,
  and an array of pointers to its daughter nodes.

  If the node content is a simple numeric type (like Byte, Int16,
  Float64, and so on, then the node content pointer points to the
  value itself. Otherwise, the pointer should be cast to the proper
  \c NodeContent pointer type: \c NodeContent_Sequence*, \c
  NodeContent_Array*, \c NodeContent_Grid*, and so on.
 */
typedef struct DAPNodeStruct {
    /*! The type of the data value indicated by this node. */
    eNODE_TYPE eType;
    /*! A pointer to the parent node in the tree.  If the node type is
       \c root, then this pointer is null. */
    struct DAPNodeStruct *nodeSuper;
    /*! A pointer to the first element in an array of pointers to the
       sub-nodes of this node.  */
    struct DAPNodeStruct **nodeSub;
    /*! A count of the number of sub-nodes. */
    unsigned int iSubNodeCount;
    /*! The data belonging to this node.  If the node type is one of the
       simple numeric types, like Byte, Int16, Float64, and so on),
       then this is a pointer to the value itself.  Otherwise, it is a
       pointer to a \c NodeContent_ structure representing the data value
       in question. */
    void *nodeContent;
    /*! The name of the variable represented by this node. */
    char *sName;
    /*! A size, in bytes, of the data on this node, and below. */
    long int nEstimatedSize;
    /*! A pointer to a list of attributes that describe this data variable. */
    DAPAttributeTable *attributes;
} DAPNode;

/*! \brief A pointer to the root node of an OPeNDAP object tree structure.

  Use this type to point to the root node of an OPeNDAP object tree
  structure.  This structure contains an indication of whether the
  structure contains a DDS, a DataDDS, a DAS or a DDX.  See \ref
  sec-data for more information about the different data objects.
*/
typedef struct {
    /*! Indicates what kind of object is stored in this tree structure. */
    eSTRUCTURE_TYPE eStructureType;
    /*! An integer containing an estimate of the size of the data
       structure, in bytes. */
    int iEstimatedSize;
    /*! The pointer to the root node. */
    DAPNode *nodeRoot;
} DAPObject;

/***************/
/** Constants **/
/***************/

#define LOGGING_None       0
#define LOGGING_NoChange  -1
#define TIMEOUT_Infinite   0
#define TIMEOUT_NoChange  -1
#define PROTOCOL_NoChange  NULL

#define VERSION_STRING "1.35"
#define VERSION_DATE "30 March 2007"
#define DEFAULT_TimeoutConnect_seconds 30
#define DEFAULT_TimeoutRead_seconds 60
#define DEFAULT_Protocol "HTTP/1.0"
#define DEFAULT_LoggingStream LOGGING_None

/*! \brief Initialize the library for statically linked code.

  When used in a statically linked context the library must be initialized
  exactly once by calling this method. When the library is used as a DLL
  or interactively the initialization will happen automatically.

  On a successful call a pointer to a ThreadSpecificStorage structure will
  be returned by reference to the client and that value must be supplied to
  certain calls in the API made by the same thread. The pointer will be null
  in case of error. The client should not allocate or attempt to free this
  pointer.
*/
eRESULT zInitializeOCAPI(ThreadSpecificStorage ** hTSS, char **sError);

/*! \brief Cleanup the library for statically linked code.

  When you are done using the library you should call this procedure.
  This procedure now free's the ThreadSpecificStorage in ocapi.c. rd 06/18/08
*/
void vCleanupOCAPI(ThreadSpecificStorage * pTSS);

/*************/
/** Utility **/
/*************/

/*! \brief Return the OCAPI version.

  Returns the official version string of the OCAPI (example: "1.05").
  The OCAPI version is composed of a cardinal version number
  plus a build number (which is at least two digits wide) separated
  by a dot.
*/
char *getVersion(void);

/*! \brief Return OCAPI version date.

  Returns the date the version was committed to CVS in the format
  YYYY Mmmm D
*/
char *getVersionDate(void);

/*! \brief Returns a description of the configuration.

  Returns a descriptive string showing all the configuration settings
  currently active.
*/
char *getConfiguration(ThreadSpecificStorage * pTSS);

/*! \brief Return documentation.

  Returns a string containing the man pages for using the OCAPI.
*/
char *getDocumentation(void);

/*! \brief Return conversion formats.

  Returns a zero-based array of \a ConversionFormats.  This feature of
  the OCAPI is not implemented.
*/
#if 0
/* not used. jhrg 4/2/2007 */
ConversionFormats *getConversionFormats(void);
#endif

/*! \brief Configures the connection parameters to be used when making data requests.

  Normally you would call this at the beginning of your program.
  You can change parameters selectively by passing the "no change"
  value for the parameters you don't want to change.

  \param iConnectTimeout_seconds The number of seconds to wait before
  timing out on a connect. defaults to 30, -1 = no change, 0 =
  infinite.

  \param iReadTimeout_seconds The number of seconds to wait before
  timing out on a data read. defaults to 60, -1 = no change, 0 =
  infinite.

  \param sProtocol The connection protocol to use. defaults to
  "HTTP/1.0", NULL or "" = no change.

  \param pLoggingStream If logging messages should be directed to a
  particular stream supply it, otherwise there will be no logging.

  \param eLogLevel The detail of logging desired. Defaults to \c
  LL_Normal.

  \param sUserAgent When an HTTP connection is made the client tells
  the server what kind of client it is using this string. Defaults to
  \c OCAPI/[version].

  \param sError If there was an error making a setting it will be
  returned to this handle.
*/
eRESULT configure(ThreadSpecificStorage * pTSS,
		  int iConnectTimeout_seconds,
		  int iReadTimeout_seconds,
		  char *sProtocol,
		  FILE * pLoggingStream,
		  /* not used jhrg 4/3/07 eLOG_LEVEL    eLogLevel, */
		  char *sUserAgent,
		  char **sError);

/*************************/
/** Proxy Server Access **/
/*************************/

/*! \brief Configures the connection parameters for a proxy host.

  \param sHost The proxy host server.

  \param iPort The port the proxy server is using.

  \param sUser The user name.

  \param sPassword The password of the user.

  \param sError If there was an error making a setting it will be
  returned to this handle.
*/
eRESULT configure_proxy(ThreadSpecificStorage * pTSS, char *sHost, int iPort,
	char *sUser, char *sPassword, char **sError);

/*************************/
/** HTTP AUTHENTICATION **/
/*************************/
/*! \brief Configures the connection parameters for http authentication.

  \param sUserC The user name provided for HTTP.

  \param sPasswordC The password provided for HTTP.

  \param sError If there was an error making a setting it will be
  returned to this handle.
*/
eRESULT configure_httpAuth(ThreadSpecificStorage * pTSS, char *sUserC,
	char *sPasswordC, char **sError);

/*************************/
/** COOKIES **/
/*************************/
/*! \brief Configures the connection parameters for cookies.

  \param cook Contains the cookies.

  \param sError If there was an error making a setting it will be
  returned to this handle.
*/
eRESULT configure_cookies(ThreadSpecificStorage * pTSS, char *cook,
	char **sError);

/*************************/
/** SSL **/
/*************************/
/*! \brief Allow for SSL configuration to be turned off.

   \param sVerify The verification variable.

  \param sError If there was an error making a setting it will be
  returned to this handle.
*/
eRESULT configure_ssl(ThreadSpecificStorage * pTSS, int sVerify, char **sError);

/*************************/
/** Compression **/
/*************************/
/*! \brief Allow for compression.

   \param sCompress The compression variable.

  \param sError If there was an error making a setting it will be
  returned to this handle.
*/
eRESULT configure_compression(ThreadSpecificStorage * pTSS, char *sCompress,
	char **sError);


/*************************/
/** Read .dodsrc file **/
/*************************/

/**\brief Allows for a .dodsrc file to be read and parsed to obtain the default authentication values.

   \param in_file The file that contains the authentication.

  \param sError If there was an error making a setting it will be
  returned to this handle.
 */
eRESULT read_dodsrc(ThreadSpecificStorage * pTSS, char *in_file, char **sError);

/*************************/
/** Write .dodsrc file **/
/*************************/

/**\brief Allows for a .dodsrc file to be created, to allow for default authentication values.

   \param out_file The file created that contains the authentication.

  \param sError If there was an error making a setting it will be
  returned to this handle.
 */
eRESULT write_dodsrc(char *out_file, char **sError);

/*************************/
/** Generic HTTP Access **/
/*************************/

/**
  \brief Parses a URL string and creates a URL structure.

  This function takes a string containing a URL, and parses it into
  its component pieces, to be packed into a \c URL structure, and
  returns a new \c URL structure.

  \param url_string The original URL to be parsed.  This will be
  something like: \c http://opendap.org/nph-hdf/data/mydata.hdf.

  Refer to the OPeNDAP User Guide for a description of the parts of
  an OPeNDAP URL.

  \param created_url Pointer to the created \c URL structure.

  \param sError Handle which will be assigned in the event of an
  error.
	eRESULT createURL( char *url_string, URL **created_url, char **sError );

	*** FOR FUTURE IMPLEMENTATION ***
*/
/*
  \brief Parses a URL string that contains a user name and password.

  This function takes a string containing a URL, and parses it into
  its component pieces, to be packed into a \c URL structure, and
  returns a new \c URL structure.

  \param pURL The original URL to be parsed.  This will be
  something like: \c "http://user:password@my.host.org/stuff.html".

  Refer to the OPeNDAP User Guide for a description of the parts of
  an OPeNDAP URL.

  \param up Pointer to the created username and password string.

  \param sURL Pointer to the created URL without the username and password.

  \param sError Handle which will be assigned in the event of an
  error.
  */
eRESULT parseURL(char *pURL, char **sUser, char **sURL, char **sPassword,
	char **sError);


/**
  \brief Gets raw information from an OPeNDAP URL.

  Gets the content pointed to by the URL supplied. This is just the
  raw content and is not parsed into an DAPNode tree structure.
  The data from the server is returned as a single null terminated
  string, so this cannot be used for getting binary data (which is
  encoded in transit).

  In practice a developer would rarely use this function, but it can
  be useful for debugging network issues.

  \param url Pointer to a \c URL structure to hold the OPeNDAP URL
  to retrieve.

  \param content Handle of the returned content, dumped into a
  single string.

  \param pTSS Pointer to the thread-specific storage object returned
  from OCAPI initialization.

  \param sError Handle which will be assigned in the event of an
  error.
*/
eRESULT getContent_byURL(URL * url, char **content,
	ThreadSpecificStorage * pTSS, char **sError);

/*! \brief Gets raw information from an OPeNDAP URL.

  Gets the content pointed to by the URL string supplied. This is
  exactly like \c GetContent_byURL, but uses a URL in a character
  string instead of a \c URL structure. The data from the server is
  returned as a single null terminated string, so this cannot be
  used for getting binary data (which is encoded in transit).

  In practice a developer would rarely use this function, but it can
  be useful for debugging network issues.

  \param url Pointer to a character string containing an OPeNDAP URL.

  \param content Handle to the string containing the raw data
  returned by the server.

  \param pTSS Pointer to the thread-specific storage object returned
  from OCAPI initialization.

  \param sError Handle which will be assigned in the event of an
  error.
*/
eRESULT getContent_byRawURL(char *sURL, char **content,
	ThreadSpecificStorage * pTSS, char **sError);

/*! \brief Gets raw information from an OPeNDAP URL.

  Gets some OPeNDAP data via the connection parameters supplied.
  This is exactly like \c GetContent_byURL, but uses a \c
  ConnectionParameters structure to describe where to find the data
  instead of a \c URL structure. The data from the
  server is returned as a single null terminated string, so this
  cannot be used for getting binary data.

  In practice a developer would rarely use this function, but it can
  be useful for debugging network issues.

  \param parameters The parameters to be used in making the
  connection.

  \param content Pointer to a string containing the content returned
  by the server.

  \param sError Handle which will be assigned in the event of an error.
*/
eRESULT getContent_byParameters(ConnectionParameters * parameters,
	char **content, char **sError);

/*! \brief Install a progress indicator.

  Sets a callback function to inform your program of the progress of
  a download.

  \param vUpdateProgress
  The callback function must return void and take three
  parameters, int \c iBytesReceived, int \c iBytesTotal and char* \c
  sStatus.

  -  The first int is the number of bytes received so far
  -  The second int is the total number of bytes expected. If
  unknown it will be zero.
  - The status string is a null-terminated string describing what
  the state of the download is. Your callback function can simply
  display it along with its progress message calculated from the
  bytes received against the bytes total.

  Pass NULL to turn off progress reporting. NULL is the default.

  There is an example of a progress reporting callback function
  called \c vUpdateProgress in \c interactive.c, provided with the
  OCAPI distribution.
*/
void setProgress(void(*vUpdateProgress)(int iBytesReceived, int iBytesTotal,
	char *sStatus));

/************************/
/** Structure Handling **/
/************************/

/*! \brief Loads an OPeNDAP structure from a URL.

  Connects to the URL and loads the returned data into an
  DAPObject if possible. If the content type is a DDS then
  each node's nodeContent member will always be null since there is
  no data in DDSes so they have no content. All other fields will be
  filled.  If the content type is a DAS then the nodeContent will be
  null and the node type will  always be NODE_Structure (except for
  the root) because in a DAS the node types are unknown.

  If an error occurs then FAILURE will be returned and the structure
  handle should not be used.

  \param url Pointer to a \c URL structure that indicates where the
  data will be coming from.

  \param structure Handle to the OPeNDAP tree structure which will be
  created for you.

  \param sError Handle which will be assigned in the event of an
  error.
*/
eRESULT structure_Load_URL(URL * url, DAPObject ** structure,
	ThreadSpecificStorage * pTSS, char **sError);

/*! \brief Loads an OPeNDAP structure from data in a stream.

  Loads an \c DAPObject from a stream.

  If an error occurs then FAILURE will be returned and the structure
  handle should not be used.

  \param pStream A pointer to the stream or file from which to read
  the data.

  \param structure Handle to an OPeNDAP structure which will be
  created for you.

  \param sError Handle which will be assigned in the event of an
  error.
eRESULT structure_Load_File( FILE *pStream, DAPObject **structure, char **sError );

	*** FOR FUTURE IMPLEMENTATION ***
*/

/*! \brief Loads an OPeNDAP structure from data in a file.

  Loads an \c DAPObject from a file.

  If an error occurs then FAILURE will be returned and the structure
  handle should not be used.

  \param sFilePath The complete path to the file. Example:
  c:\test\mydata\DrifterSeqMay06.dods
  The path syntax must be in the form recognized by the operating system being used.

  \param structure Handle to an OPeNDAP structure which will be
  created for you.

  \param sError Handle which will be assigned in the event of an
  error.
*/
eRESULT structure_Load_File(char *sFilePath, DAPObject ** structure,
	/* not used jhrg 4/3/07 ThreadSpecificStorage *pTSS, */
	char **sError);

/*! \brief Loads an OPeNDAP structure from data in a string.

  Loads an \a DAPObject from a null-terminated string, such
  as you might have been returned from \c getContent_byURL.  This
  can only be used for non-binary structures like DDSs and DASs.

  If an error occurs then FAILURE will be returned and the structure
  handle should not be used.

  \param s The input string.

  \param structure Handle to an OPeNDAP structure which will be
  created for you.

  \param sError Handle which will be assigned in the event of an
  error.
*/
eRESULT structure_Load_String(char *s, DAPObject ** structure,
	ThreadSpecificStorage * pTSS, char **sError);

/*! \brief Adds DAS data to an OPeNDAP structure.

  This function copies the attribute data in an OPeNDAP structure
  built by parsing a DAS into the attribute tables in an OPeNDAP
  structure built by parsing a DDS or DataDDS.

  If an error occurs then FAILURE will be returned. In the event of
  an error the attributes from the DAS may have been partially added
  to the target.

  \warning If the contents of \a DAS are not the attributes from the
  same data sources as the DDS in \a target, it is possible that an
  error may not result, but the data won't be correct, either.

  \param DAS Pointer to the DAS structure containing the attributes
  to be added to the target.

  \param target Pointer to the structure to which the DAS should be
  added. Must be a DDS or DataDDS.

  \param sError Handle which will be assigned in the event of an
  error.
*/
eRESULT structure_AddDAS(DAPObject * DAS, DAPObject * target, char **sError);

/*! \brief Frees memory used by an OPeNDAP structure.

  Frees memory taken by an \a DAPObject. You must delete all
  references to an object once it has been freed even if the call
  fails (because the object may be invalid).

  \param structure Handle to the OPeNDAP structure to be trashed.

  \param sError Handle which will be assigned in the event of an
  error.
*/
eRESULT structure_Free(DAPObject * structure, char **sError);

/*! \brief Clone an OPeNDAP structure.

  Creates an completely independent copy of an OPeNDAP structure.

  In the event of an error, do not use the new structure.

  \param structure The OPeNDAP structure to be cloned

  \param new_structure The output copy of \a structure.

  \param sError Handle which will be assigned in the event of an
  error.

eRESULT structure_Clone( DAPObject *structure, DAPObject **new_structure, char **sError );

	*** FOR FUTURE IMPLEMENTATION ***
*/

/*! \brief Verify an OPeNDAP structure.

  Verifies that an OPeNDAP structure is internally consistent.
  This function has already been run on any such newly-created
  structure before it is returned to you.  If you modify an
  DAPObject you may want to use this function to verify that
  it is still valid.

  \param pStructure The OPeNDAP structure to be verified

  \param sError Handle which will be assigned in the event of an error
*/
eRESULT structure_Verify(DAPObject * pStructure, char **sError);

/*! \brief Print an OPeNDAP structure.

  Prints an ASCII representation of an \c DAPObject. This
  output should exactly match the ASCII output of the same object
  from an OPeNDAP server.

  \warning This is not implemented.

  \param structure The OPeNDAP structure to be printed

  \param hBuffer Pointer to the output string.

  \param sError Handle which will be assigned in the event of an error

eRESULT structure_Print_ASCII_toString(zFetch_Asynchronous
				       DAPObject *structure,
				       char **hBuffer,
				       char **sError );

	*** FOR FUTURE IMPLEMENTATION ***

*/

/*! \brief Prints an OPeNDAP structure to a file or stream.

  Prints an ASCII representation of an \c DAPObject. This
  output should exactly match the ASCII output of the same object
  from an OPeNDAP server. Outputs the ASCII to a file or stream.

  \param structure The OPeNDAP structure to be printed.

  \param stream FILE pointer to the destination stream or file. Can
  be stdout.

  \param sError Handle which will be assigned in the event of an
  error.
*/
eRESULT structure_Print_ASCII_toFile(DAPObject * structure, FILE * stream,
	char **sError);

/*! \brief Prints an XML representation of an OPeNDAP structure.

  Prints an XML representation of the \a DAPObject. This
  output should exactly match the DDX output of the same object from
  an OPeNDAP server.

  \warning This is not implemented.

  \param structure The OPeNDAP structure to be printed.

  \param hBuffer Pointer to the output string.

  \param sError Handle which will be assigned in the event of an
  error.

eRESULT structure_Print_XML_toString( DAPObject *structure, char **hBuffer, char **sError );

	*** FOR FUTURE IMPLEMENTATION ***

*/

/*! \brief Print an XML representation of an OPeNDAP structure.

  Prints an XML representation of an \c DAPObject. This
  output should exactly match the DDX output of the same object from
  an OPeNDAP server.

  \warning This function is not implemented.

  \param structure The \c DAPObject to be printed

  \param stream FILE pointer to the destination stream or file. Can
  be stdout.

  \param sError Handle which will be assigned in the event of an
  error.

eRESULT structure_Print_XML_toFile( DAPObject *structure, FILE *stream, char **sError );

	*** FOR FUTURE IMPLEMENTATION ***

*/

/*! \brief Converts an OPeNDAP structure to a file format.

  Converts an DAPObject to another format. See \a eFORMATS
  for the available options.

  The DAPObject remains unchanged by this operation. The
  conversion may or may not be complete depending on the
  capabilities of the format.

  \warning This function is not implemented.

  \param structure The \c DAPObject to be converted

  \param eFORMAT The format into which to convert the structure.

  \param output The output string.

  \param sError Handle which will be assigned in the event of an
  error.

eRESULT structure_Convert_toString( DAPObject *structure, unsigned int eFORMAT, void **output, char **sError );

	*** FOR FUTURE IMPLEMENTATION ***

*/

/*! \brief Converts an OPeNDAP structure to a file format.

  Converts an DAPObject to another format. See \a eFORMATS
  for the available options.

  The DAPObject remains unchanged by this operation. The
  conversion may or may not be complete depending on the
  capabilities of the format.

  \warning This function is not implemented.

  \param structure The \c DAPObject to be converted

  \param eFORMAT The format into which to convert the structure.

  \param stream FILE pointer to the destination stream or file. Can
  be stdout.

  \param sError Handle which will be assigned in the event of an
  error.

eRESULT structure_Convert_toFile( DAPObject *structure, unsigned int eFORMAT, FILE *stream, char **sError );

	*** FOR FUTURE IMPLEMENTATION ***
*/

/************************/
/**     Accessors      **/
/************************/

/*! \brief Get a template node by name.

	Returns a pointer to a node if the name is supplied.

	If the name is fully qualified then the exact node will be returned.

	If the name is the name of the node only then the first node
	encountered with that name will be returned. Using the unqualified
	name is sufficient if there are not multiple nodes with the same
	name in the dataset

	If you are working with nested sequence data you will need to start at the top of
	of the sequences and work your way down. There is no way to grab an isolated row
	from a nested sequence because it will be undefined which rows in the sequences
	above. Just to make this clear: if you retrieve a node from inside a nested sequence
	the data content of that node will always be NULL. All the data is in the content
	node of the topmost sequence. See the examples and documentation for information on
	how to work down into a nested sequence.

	Always check for a NULL result. If you get a null result check your two parameters
	to see if either of them are null. If neither is NULL assume that the node name
	was not found.

	\param structure The structure (dds/datadds/das etc) containing the node desired.

	\param sNodeName The plain or fully-qualified name of the node desired. Case-sensitive.

	\return A pointer to the node or NULL if the node is not found (or the structure is invalid).
*/
DAPNode *getNodeByName(DAPObject * structure, char *sNodeName);

/*! \brief Access Sequence data in an OPeNDAP structure.

  Returns a pointer to the vector containing the data corresponding
  to a node in an \c DAPObject. This is a convenience
  function that allows the programmer to access data in a sequence
  without having to have knowledge of how that data is stored inside
  the sequence content.

  To use the accessor first traverse the sequence DataDDS structure
  to the node you are interested in and pass a pointer to the node
  to the function. The node must be a leaf node. If you pass in a
  container-type node, NULL will be returned and there will be an
  error message in sError.

	\param nodeField The leaf node corresponding to the data desired. Must
	be inside a sequence. The sequence must be a top-level sequence

	\param sError Handle which will be assigned in the event of an error

	\return A pointer to the vector. If the data is a complex data
	form such as a string, or an array then the pointer will be to a
	list of pointers to each array. There will be one pointer for
	every row in the sequence. If the data is primitive type then the
	vector will contain the actual data points themselves.
*/
void *getSequenceVector(DAPNode * nodeField, char **sError);

/*! \brief Gets the number of rows for a sequence

	\param nodeSequence The leaf node corresponding to the data desired. Must
	be inside a sequence. The sequence must be a top-level sequence

	\param sError Handle which will be assigned in the event of an error

	\return The number of rows there are for the sequence containing this field or -1 if there was an error.
*/
int getRowCount(DAPNode * nodeSequence, char **sError);

/*! \brief Gets the number of fields for a sequence

	\param nodeSequence The leaf node corresponding to the data desired. Must
	be inside a sequence. The sequence must be a top-level sequence

	\param sError Handle which will be assigned in the event of an error

	\return The number of rows there are for the sequence containing this field or -1 if there was an error.
*/
int getFieldCount(DAPNode * nodeSequence, char **sError);

/*! \brief Access Sequence data in an OPeNDAP structure.

	Same as above except for nested sequences.

	\param nodeField The leaf node corresponding to the data desired. Must
	be inside a sequence.

	\param nodeSequence The corresponding content node for the sequence.
	This must be the content node for the sequence which is the parent of
	nodeField.

	\param sError Handle which will be assigned in the event of an error

	\return A pointer to the vector. If the data is a complex data
	form such as a string, or an array then the pointer will be to a
	list of pointers to each array. There will be one pointer for
	every row in the sequence. If the data is primitive type then the
	vector will contain the actual data points themselves.
*/
void *getSequenceVector_Nested(DAPNode * nodeField,
	NodeContent_Sequence * nodeSequence, char **sError);

/*! \brief Access a data value in an OPeNDAP sequence.

	Same as the function getSequenceVector with the difference that it returns
	a list of pointers to each field in particular row of a sequence.

	\param nodeSequence The sequence containing the row desired. It must be a top-level
	sequence. If it is not top-level an error will be returned. Use getSequenceValue_Nested()
	to get values from sequences which are inside other sequences.

	\param field1 The field number desired. One-based.

	\param row1 The row number desired. One-based.

	\param sError Handle which will be assigned in the event of an error

	\return A pointer to the node content. The pointer should be cast to the content type corresponding
	to the field in question. For example, if the value is a Float64 it should be cast to double* or if
	it is an array then it should be cast to NodeContent_Array*.
*/

void *getSequenceValue(DAPNode * nodeSequence, unsigned int field1,
	unsigned int row1, char **sError);

/*! \brief Access sequence row data in an OPeNDAP nested sequence.

	Same as the function getSequenceRow with the difference that it returns
	a list of pointers to each field in particular row of a sequence.

	\param nodeSequence The sequence containing the row desired.

	\param nodeContent The corresponding content node for the sequence.
	This is required if the node is inside a nested sequence. If the
	sequence node is a top level sequence it can null. If the sequence is
	inside of another sequence then this must be present (because otherwise
	there would be no way to tell which parent row was wanted). If this is
	null and the nodeSequence supplied is inside of another sequence an
	error will be returned.

	\param field1 The column number desired. One-based.

	\param row1 The row number desired. One-based.

	\param sError Handle which will be assigned in the event of an error

	\return A pointer to a list of pointers to each member of the row. When
	the user is done with the row the caller has the responsibility of freeing
	this list. The list is 0-based and will have the same size as the number
	of fields in nodeSequence. If the field is a complex type the list member
	will be a pointer to the corresponding nodeContent of that field. If the
	field is a simple type like an int then the list member will point to the
	value.
*/

void *getSequenceValue_Nested(DAPNode * nodeSequence,
	/* not used jhrg 4/3/07 NodeContent_Sequence* nodeContent, unsigned int field1, unsigned int row1, */
	char **sError);

/*! \brief Whether a node is a primitive type or not.

	This function is useful because if a node is a primitive type and is in
	sequence then its values are stored in the value vector directly. In the
	case of complex types a pointer to the node content is stored in the value
	vector.

	\param node_type The type of node to be tested.

	\return TRUE (== 1) or FALSE (== 0)
*/
int zIsPrimitive(eNODE_TYPE node_type);
